from copy import deepcopy
import qpalm
import numpy as np
import concurrent.futures
import pytest
import os
import scipy.sparse as spa


def test_qpalm_threaded():
    valgrind = "valgrind" in os.getenv("LD_PRELOAD", "")

    settings = qpalm.Settings()
    settings.max_iter = 300
    settings.eps_abs = 1e-200
    settings.eps_rel = 0
    settings.eps_rel_in = 0
    settings.verbose = 1

    def create_solver():
        m, n = 100, 120
        data = qpalm.Data(n, m)
        rng = np.random.default_rng(seed=123)
        Q = rng.random((n, n))
        A = rng.random((m, n))
        Q = Q.T @ Q
        data.Q = spa.csc_matrix(Q)
        data.A = spa.csc_matrix(A)
        data.q = rng.random(n)
        data.bmax = rng.random(m)
        data.bmin = -np.inf * np.ones(m)
        return qpalm.Solver(data=data, settings=settings)

    shared_solver = create_solver()

    def good_experiment():
        solver = create_solver()
        solver.solve(asynchronous=True)
        return solver.info.status_val == qpalm.Info.MAX_ITER_REACHED

    def bad_experiment():
        solver = shared_solver
        solver.solve(asynchronous=True)
        return solver.info.status_val == qpalm.Info.MAX_ITER_REACHED

    def run(experiment):
        N = 4 if valgrind else 200
        with concurrent.futures.ThreadPoolExecutor(max_workers=os.cpu_count()) as pool:
            futures = (pool.submit(experiment) for _ in range(N))
            for future in concurrent.futures.as_completed(futures):
                success = future.result()
                assert success

    run(good_experiment)
    if not valgrind:
        with pytest.raises(
            RuntimeError, match=r"^Same instance of .* used in multiple threads"
        ) as e:
            run(bad_experiment)
        print(e.value)


if __name__ == "__main__":
    test_qpalm_threaded()
    print("done.")



import qpalm
import numpy as np
import concurrent.futures
from time import sleep
import scipy.sparse as spa


def test_qpalm_cancel():
    settings = qpalm.Settings()
    settings.max_iter = 20000
    settings.eps_abs = 1e-200
    settings.eps_rel = 0
    settings.eps_rel_in = 0
    settings.verbose = 1

    def create_solver():
        m, n = 100, 120
        data = qpalm.Data(n, m)
        rng = np.random.default_rng(seed=123)
        Q = rng.random((n, n))
        A = rng.random((m, n))
        Q = Q.T @ Q
        data.Q = spa.csc_matrix(Q)
        data.A = spa.csc_matrix(A)
        data.q = rng.random(n)
        data.bmax = rng.random(m)
        data.bmin = -np.inf * np.ones(m)
        return qpalm.Solver(data=data, settings=settings)

    solver = create_solver()

    def run_solver():
        solver.solve(asynchronous=True, suppress_interrupt=True)
        return solver.info.status_val

    with concurrent.futures.ThreadPoolExecutor() as pool:
        future = pool.submit(run_solver)
        sleep(0.2)
        solver.cancel()
        assert future.result() == qpalm.Info.USER_CANCELLATION


if __name__ == "__main__":
    test_qpalm_cancel()
    print("done.")



"""
C and C++ implementation of QPALM
"""
from __future__ import annotations
import numpy
import numpy.typing
import scipy.sparse
import typing
__all__: list[str] = ['Data', 'Info', 'Settings', 'Solution', 'Solver', 'build_time', 'debug']
class Data:
    A: scipy.sparse.csc_matrix
    Q: scipy.sparse.csc_matrix
    def __init__(self, n: typing.SupportsInt, m: typing.SupportsInt) -> None:
        ...
    def _get_c_data_ptr(self) -> _QPALMData:
        """
        Return a pointer to the C data struct (of type ::QPALMData).
        """
    @property
    def bmax(self) -> typing.Annotated[numpy.typing.NDArray[numpy.float64], "[m, 1]"]:
        ...
    @bmax.setter
    def bmax(self, arg1: typing.Annotated[numpy.typing.ArrayLike, numpy.float64, "[m, 1]"]) -> None:
        ...
    @property
    def bmin(self) -> typing.Annotated[numpy.typing.NDArray[numpy.float64], "[m, 1]"]:
        ...
    @bmin.setter
    def bmin(self, arg1: typing.Annotated[numpy.typing.ArrayLike, numpy.float64, "[m, 1]"]) -> None:
        ...
    @property
    def c(self) -> float:
        ...
    @c.setter
    def c(self, arg0: typing.SupportsFloat) -> None:
        ...
    @property
    def q(self) -> typing.Annotated[numpy.typing.NDArray[numpy.float64], "[m, 1]"]:
        ...
    @q.setter
    def q(self, arg1: typing.Annotated[numpy.typing.ArrayLike, numpy.float64, "[m, 1]"]) -> None:
        ...
class Info:
    DUAL_INFEASIBLE: typing.ClassVar[int] = -4
    DUAL_TERMINATED: typing.ClassVar[int] = 2
    ERROR: typing.ClassVar[int] = 0
    MAX_ITER_REACHED: typing.ClassVar[int] = -2
    PRIMAL_INFEASIBLE: typing.ClassVar[int] = -3
    SOLVED: typing.ClassVar[int] = 1
    TIME_LIMIT_REACHED: typing.ClassVar[int] = -5
    UNSOLVED: typing.ClassVar[int] = -10
    USER_CANCELLATION: typing.ClassVar[int] = -6
    status: str
    @property
    def dua2_res_norm(self) -> float:
        ...
    @dua2_res_norm.setter
    def dua2_res_norm(self, arg0: typing.SupportsFloat) -> None:
        ...
    @property
    def dua_res_norm(self) -> float:
        ...
    @dua_res_norm.setter
    def dua_res_norm(self, arg0: typing.SupportsFloat) -> None:
        ...
    @property
    def dual_objective(self) -> float:
        ...
    @dual_objective.setter
    def dual_objective(self, arg0: typing.SupportsFloat) -> None:
        ...
    @property
    def iter(self) -> int:
        ...
    @iter.setter
    def iter(self, arg0: typing.SupportsInt) -> None:
        ...
    @property
    def iter_out(self) -> int:
        ...
    @iter_out.setter
    def iter_out(self, arg0: typing.SupportsInt) -> None:
        ...
    @property
    def objective(self) -> float:
        ...
    @objective.setter
    def objective(self, arg0: typing.SupportsFloat) -> None:
        ...
    @property
    def pri_res_norm(self) -> float:
        ...
    @pri_res_norm.setter
    def pri_res_norm(self, arg0: typing.SupportsFloat) -> None:
        ...
    @property
    def run_time(self) -> float:
        ...
    @run_time.setter
    def run_time(self, arg0: typing.SupportsFloat) -> None:
        ...
    @property
    def setup_time(self) -> float:
        ...
    @setup_time.setter
    def setup_time(self, arg0: typing.SupportsFloat) -> None:
        ...
    @property
    def solve_time(self) -> float:
        ...
    @solve_time.setter
    def solve_time(self, arg0: typing.SupportsFloat) -> None:
        ...
    @property
    def status_val(self) -> int:
        ...
    @status_val.setter
    def status_val(self, arg0: typing.SupportsInt) -> None:
        ...
class Settings:
    def __init__(self) -> None:
        ...
    @property
    def delta(self) -> float:
        ...
    @delta.setter
    def delta(self, arg0: typing.SupportsFloat) -> None:
        ...
    @property
    def dual_objective_limit(self) -> float:
        ...
    @dual_objective_limit.setter
    def dual_objective_limit(self, arg0: typing.SupportsFloat) -> None:
        ...
    @property
    def enable_dual_termination(self) -> int:
        ...
    @enable_dual_termination.setter
    def enable_dual_termination(self, arg0: typing.SupportsInt) -> None:
        ...
    @property
    def eps_abs(self) -> float:
        ...
    @eps_abs.setter
    def eps_abs(self, arg0: typing.SupportsFloat) -> None:
        ...
    @property
    def eps_abs_in(self) -> float:
        ...
    @eps_abs_in.setter
    def eps_abs_in(self, arg0: typing.SupportsFloat) -> None:
        ...
    @property
    def eps_dual_inf(self) -> float:
        ...
    @eps_dual_inf.setter
    def eps_dual_inf(self, arg0: typing.SupportsFloat) -> None:
        ...
    @property
    def eps_prim_inf(self) -> float:
        ...
    @eps_prim_inf.setter
    def eps_prim_inf(self, arg0: typing.SupportsFloat) -> None:
        ...
    @property
    def eps_rel(self) -> float:
        ...
    @eps_rel.setter
    def eps_rel(self, arg0: typing.SupportsFloat) -> None:
        ...
    @property
    def eps_rel_in(self) -> float:
        ...
    @eps_rel_in.setter
    def eps_rel_in(self, arg0: typing.SupportsFloat) -> None:
        ...
    @property
    def factorization_method(self) -> int:
        ...
    @factorization_method.setter
    def factorization_method(self, arg0: typing.SupportsInt) -> None:
        ...
    @property
    def gamma_init(self) -> float:
        ...
    @gamma_init.setter
    def gamma_init(self, arg0: typing.SupportsFloat) -> None:
        ...
    @property
    def gamma_max(self) -> float:
        ...
    @gamma_max.setter
    def gamma_max(self, arg0: typing.SupportsFloat) -> None:
        ...
    @property
    def gamma_upd(self) -> float:
        ...
    @gamma_upd.setter
    def gamma_upd(self, arg0: typing.SupportsFloat) -> None:
        ...
    @property
    def inner_max_iter(self) -> int:
        ...
    @inner_max_iter.setter
    def inner_max_iter(self, arg0: typing.SupportsInt) -> None:
        ...
    @property
    def max_iter(self) -> int:
        ...
    @max_iter.setter
    def max_iter(self, arg0: typing.SupportsInt) -> None:
        ...
    @property
    def max_rank_update(self) -> int:
        ...
    @max_rank_update.setter
    def max_rank_update(self, arg0: typing.SupportsInt) -> None:
        ...
    @property
    def max_rank_update_fraction(self) -> float:
        ...
    @max_rank_update_fraction.setter
    def max_rank_update_fraction(self, arg0: typing.SupportsFloat) -> None:
        ...
    @property
    def nonconvex(self) -> int:
        ...
    @nonconvex.setter
    def nonconvex(self, arg0: typing.SupportsInt) -> None:
        ...
    @property
    def ordering(self) -> int:
        ...
    @ordering.setter
    def ordering(self, arg0: typing.SupportsInt) -> None:
        ...
    @property
    def print_iter(self) -> int:
        ...
    @print_iter.setter
    def print_iter(self, arg0: typing.SupportsInt) -> None:
        ...
    @property
    def proximal(self) -> int:
        ...
    @proximal.setter
    def proximal(self, arg0: typing.SupportsInt) -> None:
        ...
    @property
    def reset_newton_iter(self) -> int:
        ...
    @reset_newton_iter.setter
    def reset_newton_iter(self, arg0: typing.SupportsInt) -> None:
        ...
    @property
    def rho(self) -> float:
        ...
    @rho.setter
    def rho(self, arg0: typing.SupportsFloat) -> None:
        ...
    @property
    def scaling(self) -> int:
        ...
    @scaling.setter
    def scaling(self, arg0: typing.SupportsInt) -> None:
        ...
    @property
    def sigma_init(self) -> float:
        ...
    @sigma_init.setter
    def sigma_init(self, arg0: typing.SupportsFloat) -> None:
        ...
    @property
    def sigma_max(self) -> float:
        ...
    @sigma_max.setter
    def sigma_max(self, arg0: typing.SupportsFloat) -> None:
        ...
    @property
    def theta(self) -> float:
        ...
    @theta.setter
    def theta(self, arg0: typing.SupportsFloat) -> None:
        ...
    @property
    def time_limit(self) -> float:
        ...
    @time_limit.setter
    def time_limit(self, arg0: typing.SupportsFloat) -> None:
        ...
    @property
    def verbose(self) -> int:
        ...
    @verbose.setter
    def verbose(self, arg0: typing.SupportsInt) -> None:
        ...
    @property
    def warm_start(self) -> int:
        ...
    @warm_start.setter
    def warm_start(self, arg0: typing.SupportsInt) -> None:
        ...
class Solution:
    @property
    def x(self) -> typing.Annotated[numpy.typing.NDArray[numpy.float64], "[m, 1]"]:
        ...
    @property
    def y(self) -> typing.Annotated[numpy.typing.NDArray[numpy.float64], "[m, 1]"]:
        ...
class Solver:
    def __init__(self, data: Data, settings: Settings) -> None:
        ...
    def _get_c_work_ptr(self) -> _QPALMWorkspace:
        """
        Return a pointer to the C workspace struct (of type ::QPALMWorkspace).
        """
    def cancel(self) -> None:
        ...
    def solve(self, asynchronous: bool = True, suppress_interrupt: bool = False) -> None:
        ...
    def update_Q_A(self, Q_vals: typing.Annotated[numpy.typing.NDArray[numpy.float64], "[m, 1]"], A_vals: typing.Annotated[numpy.typing.NDArray[numpy.float64], "[m, 1]"]) -> None:
        ...
    def update_bounds(self, bmin: typing.Annotated[numpy.typing.NDArray[numpy.float64], "[m, 1]"] | None = None, bmax: typing.Annotated[numpy.typing.ArrayLike, numpy.float64, "[m, 1]"] | None = None) -> None:
        ...
    def update_q(self, q: typing.Annotated[numpy.typing.NDArray[numpy.float64], "[m, 1]"]) -> None:
        ...
    def update_settings(self, settings: Settings) -> None:
        ...
    def warm_start(self, x: typing.Annotated[numpy.typing.NDArray[numpy.float64], "[m, 1]"] | None = None, y: typing.Annotated[numpy.typing.NDArray[numpy.float64], "[m, 1]"] | None = None) -> None:
        ...
    @property
    def dual_inf_certificate(self) -> typing.Annotated[numpy.typing.NDArray[numpy.float64], "[m, 1]"]:
        ...
    @property
    def info(self) -> Info:
        ...
    @property
    def prim_inf_certificate(self) -> typing.Annotated[numpy.typing.NDArray[numpy.float64], "[m, 1]"]:
        ...
    @property
    def solution(self) -> Solution:
        ...
class _QPALMData:
    pass
class _QPALMWorkspace:
    pass
__version__: str = '1.2.6'
build_time: str = 'Aug 30 2025 - 19:44:47'
debug: bool = False



"""Proximal Augmented Lagrangian method for Quadratic Programs"""

__version__ = "1.2.6"

import os
import typing

def _is_truthy(s: typing.Optional[str]):
    if s is None: return False
    return not s.lower() in ('', 'false', 'no', 'off', '0')

if not typing.TYPE_CHECKING and _is_truthy(os.getenv('QPALM_PYTHON_DEBUG')):
    from qpalm._qpalm_d import *
    from qpalm._qpalm_d import __version__ as c_version
else:
    from qpalm._qpalm import *
    from qpalm._qpalm import __version__ as c_version
assert __version__ == c_version

del _is_truthy, typing, os


import numpy as np
import scipy.sparse as sp
import qpalm
import pytest


def test_simple_3x4():
    data = qpalm.Data(3, 4)

    row = np.array([0, 0, 1, 1])
    col = np.array([0, 1, 0, 1])
    val = np.array([1, -1, -1, 2])
    data.Q = sp.csc_matrix((val, (row, col)), shape=(3, 3))

    data.q = np.array([-2, -6, 1])
    data.c = 0
    data.bmin = np.array([0.5, -10, -10, -10])
    data.bmax = np.array([0.5, 10, 10, 10])

    row = np.array([0, 1, 0, 2, 0, 3])
    col = np.array([0, 0, 1, 1, 2, 2])
    val = np.array([1, 1, 1, 1, 1, 1])
    data.A = sp.csc_matrix((val, (row, col)), shape=(4, 3))

    settings = qpalm.Settings()
    solver = qpalm.Solver(data, settings)
    solver.solve()
    sol_x = solver.solution.x
    print(sol_x, solver.info.iter)
    tol = 1e-3
    assert abs(sol_x[0] - 5.5) < tol
    assert abs(sol_x[1] - 5.0) < tol
    assert abs(sol_x[2] - (-10)) < tol

    solver.warm_start(solver.solution.x, solver.solution.y)
    solver.solve()
    print(sol_x, solver.info.iter)


def test_solution_lifetime():
    def scope():
        data = qpalm.Data(3, 4)
        row = np.array([0, 0, 1, 1])
        col = np.array([0, 1, 0, 1])
        val = np.array([1, -1, -1, 2])
        data.Q = sp.csc_matrix((val, (row, col)), shape=(3, 3))

        data.q = np.array([-2, -6, 1])
        data.c = 0
        data.bmin = np.array([0.5, -10, -10, -10])
        data.bmax = np.array([0.5, 10, 10, 10])

        row = np.array([0, 1, 0, 2, 0, 3])
        col = np.array([0, 0, 1, 1, 2, 2])
        val = np.array([1, 1, 1, 1, 1, 1])
        data.A = sp.csc_matrix((val, (row, col)), shape=(4, 3))

        settings = qpalm.Settings()
        solver = qpalm.Solver(data, settings)
        solver.solve()
        sol_x = solver.solution.x
        return sol_x

    sol_x = scope()
    print(sol_x)
    tol = 1e-3
    assert abs(sol_x[0] - 5.5) < tol
    assert abs(sol_x[1] - 5.0) < tol
    assert abs(sol_x[2] - (-10)) < tol


def test_data_element_access():
    data = qpalm.Data(3, 4)
    row = np.array([0, 0, 1, 1])
    col = np.array([0, 1, 0, 1])
    val = np.array([1, -1, -1, 2])
    data.Q = sp.csc_matrix((val, (row, col)), shape=(3, 3))

    data.q = np.array([-2, -6, 1])
    data.c = 0
    data.bmin = np.array([0.5, -10, -11, -12])
    data.bmax = np.array([1.5, 10, 11, 12])

    row = np.array([0, 1, 0, 2, 0, 3])
    col = np.array([0, 0, 1, 1, 2, 2])
    val = np.array([1, 1, 1, 1, 1, 1])
    data.A = sp.csc_matrix((val, (row, col)), shape=(4, 3))

    assert data.q.flags.owndata == False
    assert data.q.flags.writeable == True

    assert np.all(data.q == np.array([-2, -6, 1]))
    data.q[2] = 7
    assert data.q[2] == 7
    assert np.all(data.q == np.array([-2, -6, 7]))

    assert data.bmin.flags.owndata == False
    assert data.bmin.flags.writeable == True

    assert np.all(data.bmin == np.array([0.5, -10, -11, -12]))
    data.bmin[3] = -99
    assert data.bmin[3] == -99
    assert np.all(data.bmin == np.array([0.5, -10, -11, -99]))

    assert data.bmax.flags.owndata == False
    assert data.bmax.flags.writeable == True

    assert np.all(data.bmax == np.array([1.5, 10, 11, 12]))
    data.bmax[1] = 99
    assert data.bmax[1] == 99
    assert np.all(data.bmax == np.array([1.5, 99, 11, 12]))


def test_data_lifetime():
    def get_q():
        data = qpalm.Data(3, 4)
        data.q = np.array([-2, -6, 1])
        return data.q

    q = get_q()
    assert np.all(q == np.array([-2, -6, 1]))
    q[1] = -7
    assert q[1] == -7
    assert np.all(q == np.array([-2, -7, 1]))


def test_invalid_bounds():
    data = qpalm.Data(3, 4)

    row = np.array([0, 0, 1, 1])
    col = np.array([0, 1, 0, 1])
    val = np.array([1, -1, -1, 2])
    data.Q = sp.csc_matrix((val, (row, col)), shape=(3, 3))

    data.q = np.array([-2, -6, 1])
    data.c = 0
    data.bmin = np.array([0.5, -10, -10, -10])
    data.bmax = np.array([-0.5, 10, 10, 10])

    row = np.array([0, 1, 0, 2, 0, 3])
    col = np.array([0, 0, 1, 1, 2, 2])
    val = np.array([1, 1, 1, 1, 1, 1])
    data.A = sp.csc_matrix((val, (row, col)), shape=(4, 3))

    settings = qpalm.Settings()
    with pytest.raises(
        ValueError, match="^Solver initialization using qpalm_setup failed"
    ):
        qpalm.Solver(data, settings)


def test_invalid_settings():
    data = qpalm.Data(3, 4)

    row = np.array([0, 0, 1, 1])
    col = np.array([0, 1, 0, 1])
    val = np.array([1, -1, -1, 2])
    data.Q = sp.csc_matrix((val, (row, col)), shape=(3, 3))

    data.q = np.array([-2, -6, 1])
    data.c = 0
    data.bmin = np.array([0.5, -10, -10, -10])
    data.bmax = np.array([0.5, 10, 10, 10])

    row = np.array([0, 1, 0, 2, 0, 3])
    col = np.array([0, 0, 1, 1, 2, 2])
    val = np.array([1, 1, 1, 1, 1, 1])
    data.A = sp.csc_matrix((val, (row, col)), shape=(4, 3))

    settings = qpalm.Settings()
    settings.max_iter = -1
    with pytest.raises(
        ValueError, match="^Solver initialization using qpalm_setup failed"
    ):
        qpalm.Solver(data, settings)



import os

from conan import ConanFile
from conan.errors import ConanInvalidConfiguration
from conan.tools.cmake import CMakeToolchain, CMake, cmake_layout
from conan.tools.build import can_run


class QPALMRecipe(ConanFile):
    name = "qpalm"
    version = "1.2.6"
    package_type = "library"

    # Optional metadata
    license = "LGPL-3.0-or-later"
    author = "Pieter P <pieter.p.dev@outlook.com>"
    url = "https://github.com/kul-optec/QPALM"
    description = "Proximal Augmented Lagrangian method for Quadratic Programs"
    topics = ("optimization", "quadratic-program", "qp", "alm")

    # Binary configuration
    settings = "os", "compiler", "build_type", "arch"
    bool_qpalm_options = {
        "with_cxx": True,
        "with_python": False,
        "with_julia": False,
        "with_fortran": False,
        "with_mex": False,
    }
    options = (
        {
            "shared": [True, False],
            "fPIC": [True, False],
        }
        | {k: [True, False] for k in bool_qpalm_options}
        | {"with_conan_python": [True, False]}
    )
    default_options = (
        {
            "shared": False,
            "fPIC": True,
        }
        | bool_qpalm_options
        | {"with_conan_python": False}
    )

    # Sources are located in the same place as this recipe, copy them to the recipe
    exports_sources = (
        "CMakeLists.txt",
        "QPALM/*",
        "LADEL/*",
        "examples/*",
        "test/*",
        "LICENSE",
        "README.md",
    )

    generators = ("CMakeDeps",)

    def requirements(self):
        self.requires("ladel/0.0.4", transitive_headers=True)
        self.test_requires("gtest/1.17.0")
        if self.options.with_cxx:
            self.requires("eigen/3.4.0", transitive_headers=True)
        if self.options.with_python:
            self.requires("pybind11/3.0.1")
            if self.options.with_conan_python:
                self.requires("tttapa-python-dev/3.13.7")

    def build_requirements(self):
        self.tool_requires("cmake/[>=4.1 <5]")

    def configure(self):
        if self.options.get_safe("with_mex"):
            self.options["ladel/*"].with_mex = True

    def config_options(self):
        if self.settings.get_safe("os") == "Windows":
            self.options.rm_safe("fPIC")

    def validate(self):
        if self.options.with_python and not self.options.with_cxx:
            msg = "Python interface requires C++. Set 'with_cxx=True'."
            raise ConanInvalidConfiguration(msg)

    def layout(self):
        cmake_layout(self)

    def generate(self):
        tc = CMakeToolchain(self)
        tc.variables["QPALM_WITH_EXAMPLES"] = False
        for k in self.bool_qpalm_options:
            value = getattr(self.options, k, None)
            if value is not None and value.value is not None:
                tc.variables["QPALM_" + k.upper()] = bool(value)
        if can_run(self):
            tc.variables["QPALM_FORCE_TEST_DISCOVERY"] = True
        tc.generate()

    def build(self):
        cmake = CMake(self)
        cmake.configure()
        cmake.build()
        cmake.test()

    def package(self):
        cmake = CMake(self)
        cmake.install()

    def package_info(self):
        self.cpp_info.set_property("cmake_find_mode", "none")
        self.cpp_info.set_property("cmake_file_name", "QPALM")
        self.cpp_info.builddirs.append(os.path.join("lib", "cmake", "QPALM"))
        if self.options.with_cxx:
            self.cpp_info.builddirs.append(os.path.join("lib", "cmake", "QPALM_cxx"))
        if self.options.with_fortran:
            self.cpp_info.builddirs.append(
                os.path.join("lib", "cmake", "QPALM_fortran")
            )
